package xyz.matj1.bp;

import mcdr.objects.impl.ObjectMocapPoseCoordsL2;
import mcdr.sequence.impl.SequenceMocapPoseCoordsL2DTW;
import mcdr.test.utils.ObjectCategoryMgmt;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;

import java.io.IOException;
import java.time.Duration;
import java.time.Instant;
import java.util.AbstractMap;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;


/**
 * Třída pro měnění rychlosti motion-capture dat pro účely testování motion words
 *
 * @author Matěj Bagar
 */
public class LinearTimeWarp {
	public static void main(String[] args) throws IOException {
		// načtení dat
		String mwDataFilepath = "data/hdm05-annotations_specific-segment80_shift16-coords_normPOS-fps12-quantized-pivots-kmedoids-350.data";
		var mocapClass = SequenceMocapPoseCoordsL2DTW.class;
		var mwCategoryMgmt = new ObjectCategoryMgmt();
		var mocapMgmt = new ObjectMgmt(mwCategoryMgmt);
		mocapMgmt.read(mocapClass, mwDataFilepath);
		var categoryObjects = convertMocap(mocapMgmt.getCategoryObjects());

		Instant speedUpSingleTimer = Instant.now();

		// zrychlení jedné sequence
		var singleSequence = categoryObjects.values().iterator().next().get(0);
		var spedupSequence = speedUpSequence(singleSequence, 2.0);

		System.out.println("time: " + Duration.between(speedUpSingleTimer, Instant.now()).toMillis() + " ms");

		Instant speedUpAllTimer = Instant.now();

		// zrychlení všech sequencí
		var categorySpedUpObjects = speedUpCategoryObjects(categoryObjects, 2.0);

		System.out.println("time: " + Duration.between(speedUpAllTimer, Instant.now()).toMillis() + " ms");
	}

	/**
	 * Převede sequence typu LocalAbstractObjet v dané struktuře do sequencí typu SequenceMocapPoseCoordsL2DTW.
	 *
	 * @param original původní struktura s pósami typu LocalAbstractObject
	 * @return nová struktura s pósami typu SequenceMocapPoseCoordsL2DTW
	 */
	private static Map<Category, List<SequenceMocapPoseCoordsL2DTW>> convertMocap(Map<Category, List<LocalAbstractObject>> original) {
		var convertedMap = new HashMap<Category, List<SequenceMocapPoseCoordsL2DTW>>();
		for (var entry: original.entrySet()) {
			List<SequenceMocapPoseCoordsL2DTW> convertedList = new ArrayList<SequenceMocapPoseCoordsL2DTW>();
			for (var object : entry.getValue()) {
				convertedList.add((SequenceMocapPoseCoordsL2DTW) object);
			}
			convertedMap.put(entry.getKey(), convertedList);
		}
		return convertedMap;
	}

	/**
	 * Tato methoda vytvoří sequenci pós, multiple× rychlejší verzi té zadané, tím, že zkopíruje některé pósy a jiné vynechá.
	 * Pokud multiple je menší než 1, výsledek je stejný jako původní sequence.
	 * Ale pokud multiple je NaN nebo není větší než 0, tato methoda vyhodí IllegalArgumentException.
	 *
	 * @param sequence původní sequence
	 * @param multiple kolikkrát rychlejší než původní sequence má být výsledná sequence
	 * @return rychlejší verze původní sequence
	 * @throws IllegalArgumentException pokud multiple je NaN nebo není větší než 0
	 */
	public static SequenceMocapPoseCoordsL2DTW speedUpSequence(SequenceMocapPoseCoordsL2DTW sequence, Double multiple) throws IllegalArgumentException {
		if (multiple <= 0 || multiple.isNaN()) {
			throw new IllegalArgumentException("The multiple for speed is not greater than 0.");
		}

		var spedUpSequenceList = new ArrayList<ObjectMocapPoseCoordsL2>();
		var counter = 0;

		for (var item : sequence.getObjects()) {
			if (Math.floor(counter % multiple) == 0) {
				spedUpSequenceList.add(item);
			}
			counter++;
		}

		return new SequenceMocapPoseCoordsL2DTW(spedUpSequenceList);
	}

	/**
	 * Tato methoda vytvoří mapu kategorií a seznamů sequencí pós, kde každá pósa je multiple× rychlejší než ta odpovídající v zadané mapě.
	 * Pokud multiple je menší než 1, výsledek je stejný jako původní mapa.
	 * Ale pokud multiple je NaN nebo není větší než 0, tato methoda vyhodí IllegalArgumentException.
	 *
	 * @param categoryObjects původní mapa
	 * @param multiple kolikkrát rychlejší než původní sequence má být každá sequence ve výsledné mapě
	 * @return mapa se zrychlenými sequencemi
	 * @throws IllegalArgumentException pokud multiple je NaN nebo není větší než 0
	 */
	public static Map<Category, List<SequenceMocapPoseCoordsL2DTW>> speedUpCategoryObjects(
		Map<Category, List<SequenceMocapPoseCoordsL2DTW>> categoryObjects,
		Double multiple
	) throws IllegalArgumentException {
		return categoryObjects.entrySet().parallelStream().map(
			e -> new SimpleImmutableEntry<>(
				e.getKey(),
				e.getValue().parallelStream().map(
					l -> speedUpSequence(l, multiple)
				).collect(Collectors.toList())
		)).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
	}
}
