package xyz.matj1.bp;

import mcdr.objects.impl.ObjectMotionWord;
import mcdr.sequence.SequenceMotionWords;
import messif.objects.LocalAbstractObject;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.stream.Stream;

public class Utils {
	/**
	 * Returns a string representing the given ObjectMotionWord.
	 *
	 * @param seq the sequence castable to ObjectMotionWord
	 * @return a string representing the input sequence
	 */
	public static String sequenceToString(LocalAbstractObject seq) {
		StringBuilder rtv = new StringBuilder();
		for (ObjectMotionWord mw : ((SequenceMotionWords<? extends ObjectMotionWord>)seq).getSequenceData()) {
			if (rtv.length() > 0) {
				rtv.append(',');
			}
			rtv.append(mw.toString());
		}
		return seq.getLocatorURI() + " [" + rtv + "]";
	}

	/**
	 * Calculates the distance of two categories as the average of the distances of objects in each tuples from the Cartesian product of given collections.
	 *
	 * @param objectsInCategory1 a collection of the objects in the first category
	 * @param objectsInCategory2 a collection of the objects in the second category
	 * @return the distance of of the categories
	 */
	public static Double distanceOfCategories(
		Collection<LocalAbstractObject> objectsInCategory1,
		Collection<LocalAbstractObject> objectsInCategory2
	) {
		ArrayList<Float> objectDistances = new ArrayList<>();
		for (LocalAbstractObject object1 : objectsInCategory1) {
			for (LocalAbstractObject object2 : objectsInCategory2) {
				objectDistances.add(object1.getDistance(object2));
		}}
		return averageOfFloats(objectDistances);
	}

	/**
	 * Returns the average of all the values in the input collection.
	 * @param collection the input collection
	 * @return average of the values in the input collection
	 */
	public static Double average(Collection<Double> collection) {
		return collection.parallelStream().mapToDouble(x->x).average().orElse(Double.NaN);
	}

	/**
	 * Returns the average of all the values in the input collection.
	 * @param collection the input collection
	 * @return average of the values in the input collection
	 */
	public static Double averageOfFloats(Collection<Float> collection) {
		return collection.parallelStream().mapToDouble(x->(double)x).average().orElse(Double.NaN);
	}

	/**
	 * A pair of values
	 *
	 * @param first the first value
	 * @param second the second value
	 */
	public record Pair<A, B>(A first, B second) {
		public String toString() {
			return "(" + first.toString() + ", " + second.toString() + ")";
	}}

	/**
	 * Converts the given Iterator&lt;T&gt; to Stream&lt;T&gt; by putting each object returned from the iterator to the stream.
	 *
	 * @param iterator the given iterator
	 * @return a stream with the values taken from the iterator
	 */
	public static <T> Stream<T> iteratorToStream(Iterator<T> iterator) {
		Stream.Builder<T> builder = Stream.builder();
		iterator.forEachRemaining(builder::add);
		return builder.build().parallel();
	}

	/**
	 * Returns the length in poses of the given motion words sequence.
	 *
	 * @param object a motion words sequence
	 * @return the length of object
	 * @throws IllegalArgumentException if the given object isn't a motion words sequence
	 */
	public static int getLength(LocalAbstractObject object) throws IllegalArgumentException {
		try {
			return Integer.parseInt(object.getLocatorURI().split("_")[3]);
		} catch (IndexOutOfBoundsException | NumberFormatException e) {
			throw new IllegalArgumentException("The given object probably isn't a motion words sequence", e);
	}}

	/**
	 * A representation of a relation, used for pairs of sequences
	 *
	 * DIFFERENCE – difference of the lengths
	 * QUOTIENT – quotient of the lengths
	 */
	public enum LengthRelation {
		DIFFERENCE,
		QUOTIENT,
	}

	public static double normaliseQuotient(double value) throws IllegalArgumentException {
		if (!Double.isFinite(value) || value == 0d) {
			throw new IllegalArgumentException("The value to be normalised isn't non-zero finite number");
		}

		if (Math.abs(value) < 1) {
			return 1/value;
		} else {
			return value;
		}
	}
}
